// Copyright (C) 2005 Alexander Ushakov
// Contents: Definition of class ThRightNormalForm
//
// Principal Authors: Alexander Ushakov
//
// Revision History:
//

#ifndef _ThRightNormalForm_h_
#define _ThRightNormalForm_h_

#include "tuples.h"
#include "Permutation.h"

#include "vector"
#include "list"
using namespace std;


class BraidGroup;
class Word;


//---------------------------------------------------------------------------//
//--------------------------- ThRightNormalForm -----------------------------//
//---------------------------------------------------------------------------//

//! Defines a representation of a right Garside normal form//

class ThRightNormalForm
{
public:
  //! Presentation of a normal form. 
  /*!
    The first component specifies the rank of a braid group.
    The second component specifies  the power of the half twist.
    The third component specifies the list of braid permutations.
  */
  typedef triple< int , int , list< Permutation > > NF;
  
  /////////////////////////////////////////////////////////
  //                                                     //
  //  Constructors:                                      //
  //                                                     //
  /////////////////////////////////////////////////////////

public:
  
  //! Default constructor (rank specifies the rank of a braid group the form belongs to)
  /*!
    Default rank value is for map<...> only
    make sure to provide the argument for this constructor
  */
  ThRightNormalForm( int rank=0 ) : 
    theRank( rank ) ,
    theOmegaPower( 0 ) ,
    theDecomposition( ) { }
    
    
  //! Constructor (rank specifies the rank of a braid group the form belongs to, p - a power of a half twist, and d - a list of permutations. So the pair (p,d) is a presentation)
  /*!
    There is no check that the pair (p,d) defines a correct normal form representation.
    If you are not sure if (p,d) is correct apply static function adjustDecomposition first.
  */
  ThRightNormalForm( int rank , int p , const list< Permutation >& d ) : 
    theRank( rank ),
    theOmegaPower( p ) ,
    theDecomposition( d ) { }

  //! Constructor (tr specifies a presentation of the normal form)
  ThRightNormalForm( const NF& tr ) : 
    theRank( tr.first ),
    theOmegaPower( tr.second ) ,
    theDecomposition( tr.third ) { }

  //! Constructs the normal form of a braid word w.
  ThRightNormalForm( const BraidGroup& G , const Word& w );

  // Assignment operator
  ThRightNormalForm& operator = ( const NF& tr ) {
		theRank = tr.first;
    theOmegaPower = tr.second;
    theDecomposition = tr.third;
    return *this;
  }
  
  /////////////////////////////////////////////////////////
  //                                                     //
  //  Operators                                          //
  //                                                     //
  /////////////////////////////////////////////////////////
public:

  //! Cast operator (returns a representation of a normal form)
  operator NF( ) const {
    return NF( theRank , theOmegaPower , theDecomposition );
  }

  //! Compare (check if equal)
  bool operator == ( const ThRightNormalForm& rep ) const {
    return 
      theRank==rep.theRank && 
      theOmegaPower==rep.theOmegaPower && 
      theDecomposition==rep.theDecomposition;
  }
  
  //! Compare (check if not equal)
  bool operator != ( const ThRightNormalForm& rep ) const {
    return 
      theRank!=rep.theRank ||
      theOmegaPower!=rep.theOmegaPower || 
      theDecomposition!=rep.theDecomposition;
  }
  
  //! Compare (check if less, performed componentwise)
  bool operator < ( const ThRightNormalForm& rep ) const {
    if( theRank<rep.theRank )
      return true;
    if( theRank>rep.theRank )
      return false;
    if( theOmegaPower<rep.theOmegaPower )
      return true;
    if( theOmegaPower>rep.theOmegaPower )
      return false;
    return theDecomposition<rep.theDecomposition;
  }

  // Invert a normal form.  
  ThRightNormalForm operator - ( ) const {
    return inverse( );
  }
  
  //! Multiply two normal forms
  ThRightNormalForm operator * ( const ThRightNormalForm& rep ) const {
    return multiply( rep );
  }
  
  //! Multiply a normal form by the other on the right
  ThRightNormalForm& operator *= ( const ThRightNormalForm& rep ) {
    *this = multiply( rep );
    return *this;
  }

  /////////////////////////////////////////////////////////
  //                                                     //
  //  Accessors:                                         //
  //                                                     //
  /////////////////////////////////////////////////////////

public:
  
  //! Get the rank of a corresponding braid group
  inline int getRank ( ) const { return theRank; }
  //! Get half-twist power.
  inline int getPower( ) const { return theOmegaPower; }
  //! Get a list of permutations.
  inline const list< Permutation >& getDecomposition( ) const { return theDecomposition; }
  
  //! Construct a half-twist permutation
  static Permutation getHalfTwistPermutation( int theIndex ) {
    Permutation result( theIndex );
    for( int i=0 ; i<theIndex ; ++i )
      result[i] = theIndex-i-1;
    return result;
  }

  //! Check if a normal form is trivial
  bool isTrivial( ) const { 
    return theOmegaPower==0 && theDecomposition.size( )==0;
  }

  //! Computes a word represented by a normal form.
  /*!
    For each permutation-braid computes a braid-word it represents
    and, finally, concatenate all the results.
   */
  Word getWord( ) const;
  //! Computes a "shorter" word represented by a normal form.
  /*!
    Try to use the fact that the square of a half twist generates a center.
    If the power of half-twist is negative then it cancel them with the other
    permutation braids. After that, for newly obtained permutations computes
    braid words and concatenates them.
   */
  Word getShortWord( ) const;

  //! Check whether two braids are conjugate
  /*!
    Very slow function. Checks if super summit sets of two braids contain the same element (i.e., coincide).
    ... and the super summit sets are typically huge even for decent parameters.
    @return - a pair (R,C), where R is a boolean value (true if braids are conjugate, false otherwise), and C is a conjugator (if braids conjugate)
   */
  pair<bool,ThRightNormalForm>
    areConjugate( const ThRightNormalForm& rep ) const;
  
  //! Adjust a normal form
  /*!
    Use this function if the triple of arguments does not satisfy "greedy conditions".
   */
  static void adjustDecomposition( int rank , int& power , list<Permutation>& decomp );

  //! Adjust a normal form
  /*!
    Use this function if there is a possiblitiy that the sequence of permutations does not 
    satisfy the "greedy conditions". Right now this might happen only after use of a 
    constructor ThRightNormalForm (const NF &tr) or, more likely, in function setDecomposition(...). 
    This function uses adjustDecomposition( ... ).
  */
  void adjust( ) {
    adjustDecomposition( theRank , theOmegaPower , theDecomposition );
  }

  //! Compute generators of the centralizer of an element
  /*! Very slow function. Function computes the graph with the super summit set 
    of the element as the vertex set. Even for a small rank of a braid group (say 10)
    and a short braid word (say 100) the super summit set is typically huge.  
   */
  set<ThRightNormalForm> computeNormalizer( ) const;
  
  //! Find an element of a super summit set.
  /*!
    Function uses cycling and decycling to decrease the canonical length of an element.
   */
  ThRightNormalForm 
    findSSSRepresentative( ThRightNormalForm& conjugator ) const;
  ThRightNormalForm findSSSRepresentative( ) const {
    return findSSSRepresentative( theRank , *this ).first;
  }


  //! Cycle a normal form.
  /*!
    Operation: \f$\Delta^s \xi_1 \ldots \xi_{t-1} \xi_t ~\mapsto~ \xi_t \Delta^s \xi_1 \ldots \xi_{t-1}\f$.
    @return a pair (R,C), where R is the result of a cycling and C is a conjugator (if A is an input then \f$R = C^{-1} A C\f$)
   */
  static pair<NF,ThRightNormalForm> cycle( const NF& rep );
  //! Cycle a normal form.
  /*!
    Operation: \f$\Delta^s \xi_1 \xi_2 \ldots \xi_t ~\mapsto~ \Delta^s \xi_2 \ldots \xi_t \cdot (\Delta^s \xi_1 \Delta^{-s})\f$.
    @return a pair (R,C), where R is the result of a cycling and C is a conjugator (if A is an input then \f$R = C^{-1} A C\f$)
   */
  static pair<NF,ThRightNormalForm> decycle( const NF& rep );
  
  
  /////////////////////////////////////////////////////////
  //                                                     //
  //  Modifiers:                                         //
  //                                                     //
  /////////////////////////////////////////////////////////
public:

  //! Set a power of a half-twist.
  inline void setPower( int p ) { theOmegaPower = p; }
  //! Set a decomposition of a normal form (without any check of "greedy conditions"). 
  inline void setDecomposition( const list< Permutation >& d ) { theDecomposition = d; }
  
  /////////////////////////////////////////////////////////
  //                                                     //
  //  Internal functions:                                //
  //                                                     //
  /////////////////////////////////////////////////////////
private:

  //! (Low level function) Invert a normal form.
  NF inverse( ) const;
  //! (Low level function) Multiply normal forms.
  NF multiply( const ThRightNormalForm& rep ) const;

  //! (Low level function) Compute super summit set representative.
  static pair<NF,ThRightNormalForm> 
    findSSSRepresentative( int rank , const NF& rep );

  enum transformationResult { TWO_MULTIPLIERS , ONE_MULTIPLIER , NO_CHANGE };
  static transformationResult transform ( int theIndex , Permutation& p1 , Permutation& p2 );
  
  static set<Permutation> getSimpleConjugators( int rank , const NF& rep );
  
  static set<Permutation> getSimpleSummitConjugators( int rank , const NF& rep );
  
  static pair<bool,ThRightNormalForm>
    sssConstructionIteration( int rank ,
			      map< NF , NF >& sss_new1 ,
			      map< NF , NF >& sss_checked1 ,
			      const map< NF , NF >& sss_new2 ,
			      const map< NF , NF >& sss_checked2 );
  
  /////////////////////////////////////////////////////////
  //                                                     //
  //  Data members:                                      //
  //                                                     //
  /////////////////////////////////////////////////////////

private:

  //! The rank of the braid group (number of strands).
  int theRank;
  //! Power of omega.
  int theOmegaPower;
  //! Sequence of permutations.
  list< Permutation > theDecomposition;
  
};


ostream& operator << ( ostream& os, const ThRightNormalForm& nf );



#endif
