// Copyright (C) 2005 Alexander Ushakov
// Contents: Implementation of class ThRightNormalForm
//
// Principal Authors: Alexander Ushakov
//
// Revision History:
//

#include "fstream"
#include "ShortBraidForm.h"


#include "ThRightNormalForm.h"
#include "BraidGroup.h"
#include "Word.h"


ostream& printOn( ostream& os , const list<Permutation>& lp  )
{
  os << "(";
  list<Permutation>::const_iterator lp_it = lp.begin( );
  for( ; lp_it!=lp.end( ) ; ++lp_it ) {
    // os << (*lp_it) << " -> " << (*lp_it).geodesic( ).size( ) << endl;
    if( lp_it!=lp.begin( ) )
      os << ",";
    os.width(2);
    os << (*lp_it).geodesic( ).size( );
  }
  os << ")";
  
  return os;
}



ostream& operator << ( ostream& os, const ThRightNormalForm& rep )
{
  const list< Permutation >& decomposition = rep.getDecomposition( );

  list< Permutation >::const_iterator it = decomposition.begin( );
  for( ; it!=decomposition.end( ) ; ++it )
    os << *it << endl;
  
  os << "Power = " << rep.getPower( ) << endl;
	os << "Rank = " << rep.getRank( ) << endl;
  return os;
}


//---------------------------------------------------------------------------//
//--------------------------- ThRightNormalForm -----------------------------//
//---------------------------------------------------------------------------//


ThRightNormalForm::ThRightNormalForm( const BraidGroup& G , const Word& w ) :
	theRank( G.getRank( ) ),
	theOmegaPower( 0 )
{
  const Permutation omega = getHalfTwistPermutation( theRank );

  // 1. compute permutation decomposition of a given braid word
  Permutation curMult( theRank );
  bool trivialMult = true;
  for( Word::const_iterator w_it=w.end( ) ; w_it!=w.begin( ) ; ) {

		// find the current block
		Word::const_iterator w_it2 = --w_it;
		for( ; (*w_it2<0)==(*w_it<0) && w_it2!=w.begin( ) ; --w_it2 );
		if( (*w_it2<0)!=(*w_it<0) ) ++w_it2;
		w_it++;

		// cout << " ***** " << *w_it2 << endl;
		// check the sign of the block
    if( *w_it2>0 ) {

			// process the positive block
			for( Word::const_iterator w_it3=w_it ; w_it3!=w_it2 ; ) {

				--w_it3;
				int gen = *w_it3;
				// cout << " +++++ " << gen << endl;
				if( curMult[gen-1]>curMult[gen] ) {
					theDecomposition.push_front( curMult );
					curMult = Permutation( theRank );
				}
				swap( curMult[gen-1] , curMult[gen] );
				trivialMult = false;
      }
      theDecomposition.push_front( curMult );
      curMult = Permutation( theRank );
      trivialMult = true;

    } else {
      
			// process the negative block
      list< Permutation > mult;
      for( Word::const_iterator w_it3 =w_it2 ; w_it3!=w_it ; ++w_it3 ) {

				int gen = -*w_it3;
				// cout << " ------ " << gen << endl;
				if( curMult[gen-1]>curMult[gen] ) {
					mult.push_front( curMult );
					curMult = Permutation( theRank );
				}
				swap( curMult[gen-1] , curMult[gen] );
				trivialMult = false;
      }
      mult.push_front( curMult );
      curMult = Permutation( theRank );
      trivialMult = true;
			
      list< Permutation >::iterator it = mult.begin( );
      for( ; it!=mult.end( ) ; ++it ) {
				theDecomposition.push_front( omega );
				if( !((*it).inverse( )*omega).isTrivial( ) )
					theDecomposition.push_front( (*it).inverse( )*omega );
				theOmegaPower -= 2;
      }
    }

		w_it = w_it2;
  }
	
  adjustDecomposition( theRank , theOmegaPower , theDecomposition );
}


//---------------------------------------------------------------------------//
//--------------------------- adjustDecomposition ---------------------------//
//---------------------------------------------------------------------------//


void 
ThRightNormalForm::adjustDecomposition
( int rank , int& power , list<Permutation>& decomp )
{
  const Permutation omega = getHalfTwistPermutation( rank );

  bool flip = false;
  list< Permutation>::iterator it1 =  decomp.begin( );
  for( ; it1!=decomp.end( ) ; it1++ ) {
    
    if( flip )
      *it1 = omega*(*it1)*omega;
		
    list< Permutation >::iterator it2 = it1;
    for( ; it2!=decomp.begin( ) ; it2-- ) {
      list< Permutation >::iterator it3 = it2;
      it3--;
      
      transformationResult tr_res = transform( rank , *it3 , *it2 );
      switch( tr_res ) {
      case ONE_MULTIPLIER:
				if( it1==it2 )
					it1 = it2 = decomp.erase( it3 );
				else
					it2 = decomp.erase( it3 );
				it2++;
				break;
      case NO_CHANGE:
				it2 = decomp.begin( );
				it2++;
				break;
      }
    }
		
    if( *it1==omega ) {
      power++;
      it1 = decomp.erase( it1 );
      it1--;
      flip = !flip;
    }
		
  }
}


//---------------------------------------------------------------------------//
//------------------------------- transform ---------------------------------//
//---------------------------------------------------------------------------//


// finds longest head of p2 
// that can be multiplied by p1 on the left
ThRightNormalForm::transformationResult
ThRightNormalForm::transform( int theRank , Permutation& p1 , Permutation& p2 )
{
  transformationResult res = TWO_MULTIPLIERS;
  const Permutation omega = getHalfTwistPermutation( theRank );

  Permutation p3 = p1.meet( omega * -p2 );
  if( p3==p1 )
    res = ONE_MULTIPLIER;
  if( p3.isTrivial( ) )
    res = NO_CHANGE;
  p2 = p3 * p2;
  p1 *= -p3;

  return res;
}


//---------------------------------------------------------------------------//
//------------------------------- inverse -----------------------------------//
//---------------------------------------------------------------------------//


ThRightNormalForm::NF
ThRightNormalForm::inverse( ) const
{
  int dec_size = theDecomposition.size( );
  
  int power = -theOmegaPower-dec_size;
  const Permutation omega = getHalfTwistPermutation( theRank );
  list<Permutation> result;
  
  list<Permutation>::const_iterator it = --theDecomposition.end( );
  for( int i=0 ; i<dec_size ; ++i , --it )
    if( ( i%2==0 )==( theOmegaPower%2!=0 ) )
      result.push_back( omega * -(*it) );
    else
      result.push_back( -(*it) * omega );
  
  return NF( theRank , power , result );
}


//---------------------------------------------------------------------------//
//------------------------------- multiply ----------------------------------//
//---------------------------------------------------------------------------//


ThRightNormalForm::NF
ThRightNormalForm::multiply( const ThRightNormalForm& rep ) const
{
  const Permutation omega = getHalfTwistPermutation( theRank );
  
  // 1. shift omegas to the right
  int power = theOmegaPower + rep.theOmegaPower;
  
  list< Permutation > blocks( theDecomposition );
  
  // 2. if power of the omega on the left is odd we must flip 
  //    permutations of the second form
  list< Permutation >::const_iterator it = rep.theDecomposition.begin( );
  if( theOmegaPower%2 ) {
    for( int t=0 ; t<rep.theDecomposition.size( ) ; ++t , ++it )
      blocks.push_back( omega * (*it) * omega );
  } else
    blocks.insert( blocks.end( ) , 
		   rep.theDecomposition.begin( ) , 
		   rep.theDecomposition.end( ) );
  
  adjustDecomposition( theRank , power , blocks );
  return NF( theRank , power , blocks );
}


//---------------------------------------------------------------------------//
//-------------------------------- getWord ----------------------------------//
//---------------------------------------------------------------------------//


Word 
ThRightNormalForm::getWord( ) const
{
  int i;

  Word result;
  vector< int > decomposition;
  
  vector< int > geodesic;
  list<Permutation>::const_iterator it = theDecomposition.begin( );
  for( ; it!=theDecomposition.end( ) ; ++it ) {
    geodesic = (*it).geodesic( );
    for( int j=0 ; j<geodesic.size( ) ; ++j )
      result.push_back( geodesic[j]+1 );
  }

  if( theOmegaPower==0 )
    return result;
  
  Word omegaWord;
  const Permutation omega = getHalfTwistPermutation( theRank );
  geodesic = omega.geodesic( );
  for( i=0 ; i<geodesic.size( ) ; ++i )
    omegaWord.push_back( geodesic[i]+1 );
  
  if( theOmegaPower<0 )
    omegaWord = -omegaWord;
    
  for( i=0 ; i<abs(theOmegaPower) ; ++i )
    result *= omegaWord;
  
  return result;
}


Word ThRightNormalForm::getShortWord( ) const
{
  Word result;
  Permutation omega = getHalfTwistPermutation( getRank( ) );
	int power = getPower( );

  if( power<0 ) {

    const list< Permutation >& decomp = getDecomposition( );
    list<Permutation>:: const_iterator it = decomp.begin( );
    for( int j=0 ; it!=decomp.end( ) ; ++it, ++j ) {
      int n = j - decomp.size( ) - power;
      if( n<0 ) {
				vector< int > gd = (*it).geodesic();
				for( int t=0 ; t<gd.size() ; ++t )
					result.push_back( gd[t]+1 );
      } else {
				
				Permutation p = ( n%2 == 1 ? (*it).inverse() * omega : omega * (*it).inverse() );
				
				vector<int> gd = p.geodesic();
				for( int t=gd.size( )-1 ; t>=0 ; --t )
					result.push_back( -gd[t]-1 );
      }
    }
  } else
    result = getWord( );
  
  return result;
}

//---------------------------------------------------------------------------//
//------------------------------ cycle/decycle ------------------------------//
//---------------------------------------------------------------------------//


pair< ThRightNormalForm::NF , ThRightNormalForm > 
ThRightNormalForm::cycle( const NF& rep )
{
  ThRightNormalForm conjugator( rep.first );

  // if decomposition is empty
  if( rep.third.begin()==rep.third.end() )
    return pair<NF,ThRightNormalForm> ( rep , conjugator );
  
  int theRank = rep.first;
  const Permutation omega = getHalfTwistPermutation( theRank );
  
  NF pr = rep;
  Permutation conj;
  if( rep.second%2==0 )
    conj = *(--pr.third.end( ));
  else
    conj = omega * (*(--pr.third.end( ))) * omega;
  pr.third.push_front( conj );
  pr.third.pop_back( );
  
  conj = conj.inverse( ) * omega;
  conjugator = ThRightNormalForm( theRank , -1 , list<Permutation>(1,conj) );
  
  adjustDecomposition( pr.first , pr.second , pr.third );
  return pair<NF,ThRightNormalForm> ( pr , conjugator );
}


pair< ThRightNormalForm::NF , ThRightNormalForm > 
ThRightNormalForm::decycle( const NF& rep )
{
  ThRightNormalForm conjugator( rep.first );

  if( rep.third.begin()==rep.third.end() )
    return pair<NF,ThRightNormalForm> ( rep , conjugator );
  
  int theRank = rep.first;
  const Permutation omega = getHalfTwistPermutation( theRank );
  
  NF pr = rep;
  Permutation conj;
  if( rep.second%2==0 )
    conj = *pr.third.begin( );
  else 
    conj = omega * (*pr.third.begin( )) * omega;
  pr.third.push_back( conj );
  pr.third.pop_front( );
  
  conjugator = 
    ThRightNormalForm( theRank , 0 , list<Permutation>(1,*rep.third.begin( )) );
  
  adjustDecomposition( theRank , pr.second , pr.third );
	// cout << "++++++++++++++++++++++++++++++++++" << endl;
	// ThRightNormalForm a = rep, b = pr, c = conjugator;
	// cout << (b == -c * a * c) << endl;
	// cout << "++++++++++++++++++++++++++++++++++" << endl;
  return pair<NF,ThRightNormalForm> ( pr , conjugator );
}


//---------------------------------------------------------------------------//
//--------------------------- findSSSRepresentative -------------------------//
//---------------------------------------------------------------------------//


pair< ThRightNormalForm::NF , ThRightNormalForm > 
ThRightNormalForm::findSSSRepresentative
( int rank , const NF& rep )
{
  bool progress;
  
  NF cur_nf( rep );
  NF result( rep );
  ThRightNormalForm cur_conj( rank );
  ThRightNormalForm conjugator( rank );
  for( progress=true ; progress ; ) {
    progress = false;
    for( int i=0; i<rank ; ++i ) {
      pair<NF,ThRightNormalForm> pr = cycle( cur_nf );
      cur_nf = pr.first;
      cur_conj *= pr.second;
/*
      {
	ThRightNormalForm tmp2 = -cur_conj * ThRightNormalForm(rep) * cur_conj;
	if( tmp2!=cur_nf ) {
	  cout << "rep = " << ThRightNormalForm(rep) << endl;
	  cout << "cur_nf = " << ThRightNormalForm(cur_nf) << endl;
	  cout << "tmp2 = " << tmp2 << endl;
	  cout << "cur_conj = " << cur_conj << endl;
	  cout << "Doesn't work!!!" << endl;
	  exit(1);
	}
      }
*/
      if( cur_nf.first>result.first ) {
        result = cur_nf;
        conjugator = cur_conj;
        progress = true;
        break;
      }
    }
  }
  
  cur_nf = result;
  cur_conj = conjugator;
  for( progress=true ; progress ; ) {
    progress = false;
    for( int i=0; i<rank ; ++i ) {
      pair<NF,ThRightNormalForm> pr = decycle( cur_nf );
      cur_nf = pr.first;
      cur_conj *= pr.second;
/*
      {
	ThRightNormalForm tmp2 = -cur_conj * ThRightNormalForm(rep) * cur_conj;
	if( tmp2!=cur_nf ) {
	  cout << "rep = " << ThRightNormalForm(rep) << endl;
	  cout << "cur_nf = " << ThRightNormalForm(cur_nf) << endl;
	  cout << "tmp2 = " << tmp2 << endl;
	  cout << "cur_conj = " << cur_conj << endl;
	  cout << "Doesn't work***" << endl;
	  exit(1);
	}
      }
*/
      if( cur_nf.third.size( )<result.third.size( ) ) {
				result = cur_nf;
				conjugator = cur_conj;
				progress = true;
				break;
      }
    }
  }
  
  return pair<NF,ThRightNormalForm>( result , conjugator );
}


//---------------------------------------------------------------------------//
//-------------------------- findSSSRepresentative --------------------------//
//---------------------------------------------------------------------------//


ThRightNormalForm ThRightNormalForm::findSSSRepresentative
( ThRightNormalForm& conjugator ) const 
{
  typedef triple< int , int , list<Permutation> > NF;
  
  pair<NF,ThRightNormalForm> pr = ThRightNormalForm::findSSSRepresentative( theRank , *this );
  conjugator = ThRightNormalForm( pr.second );
  return pr.first;
}


//---------------------------------------------------------------------------//
//--------------------------- getSimpleConjugators --------------------------//
//---------------------------------------------------------------------------//


Permutation 
temp_join( Permutation p1 , Permutation p2 , const Permutation& omega )
{

  Permutation P3 = omega *p1.inverse( );
  Permutation P4 = omega * p2.inverse( );
  return (omega*P3.meet( P4 )).inverse( );
}


set<Permutation> 
ThRightNormalForm::getSimpleConjugators
( int rank , const NF& rep )
{
  const Permutation omega = getHalfTwistPermutation( rank );
  list< Permutation > decomp = rep.third;
  
  set<Permutation> result;
  
  for( int i=0 ; i<rank-1 ; ++i ) {
    Permutation c = Permutation( rank );
    c.change( i , i+1 );
    
    bool done = false;
    if( rep.second>0 )
      done = true;
    
    Permutation v = c;
    list< Permutation >::const_iterator it = decomp.begin( );
    while( !done ) {
      for( it = decomp.begin( ) ; it!=decomp.end( ) ; ++it ) {
				// cout << v << " ^ " << *it << " -> " << temp_join( v , *it , omega ) << " , " << (*it).inverse() * temp_join( v , *it , omega ) << endl;
				v = (*it).inverse() * temp_join( v , *it , omega );
				
				if( v.isTrivial( ) ) {
					done = true;
					break;
				}
      }
      if( !done ) {
				v = temp_join( v , c , omega );
				if( v==c )
					done = true;
				else
					c = v;
      }
    }
    
    if( rep.second==0 ) {
			bool f = c==omega;
      ThRightNormalForm r_mult( rank , (f ? 1:0) , (f?list<Permutation>( 0 ) : list<Permutation>( 1 , c ) ) );
			ThRightNormalForm new_el = -r_mult * rep * r_mult;
      
      if( new_el.theOmegaPower<0 ) {
				cerr << "Problem (56920)" << endl;
				exit(2);
      }
    }
    
    result.insert( c );
  }
  
  return result;
}


//---------------------------------------------------------------------------//
//------------------------- getSimpleSummitConjugators ----------------------//
//---------------------------------------------------------------------------//
// WARNING!!! DOESN'T WORK FOR NON-SUMMIT ELEMENTS!!! (GETS INTO A LOOP)


set<Permutation> 
ThRightNormalForm::getSimpleSummitConjugators
( int rank , const NF& rep )
{
  const Permutation omega = getHalfTwistPermutation( rank );
  
  set<Permutation> result;
  
  set<Permutation> conjs = getSimpleConjugators( rank , rep );
  set<Permutation>::iterator it = conjs.begin( );
  for( ; it!=conjs.end( ) ; ++it ) {

    // cout << "*" << endl;
    bool done = false;
    
    Permutation c( *it );
    ThRightNormalForm res( rank );
    for( int counter=0 ; !done ; counter++ ) {
      
      if( c==omega )
	break;
      
      ThRightNormalForm r_mult( rank , 0 , list<Permutation>( 1 , c ) );
      ThRightNormalForm l_mult = r_mult.inverse( );

      res = l_mult.multiply( rep );
      res = res.multiply( r_mult );

      /*
      if( counter>1000 ) {
	cout << counter << " : c = " << c << endl;
	cout << "~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~" << endl;
	cout << l_mult << endl;
	cout << "~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~" << endl;
	cout << r_mult << endl;
	cout << "~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~" << endl;
	cout << (ThRightNormalForm)rep << endl;
	cout << "~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~" << endl;
	cout << res << endl;
	cout << "~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~" << endl;
      }
      */
      
      if( res.theDecomposition.size( )==rep.third.size( ) ) {
				done = true;
				continue;
      }
      if( res.theDecomposition.size( )+res.theOmegaPower > 
				rep.third.size( )+rep.second ) {
				c = c * (*res.theDecomposition.begin( ));
				continue;
      }
      
      Permutation v = *--res.theDecomposition.end( );
      if( 1-rep.second%2 )
				v = omega * v * omega;
      v = v.inverse( ) * omega;
      c = c*v;
    }
    result.insert( c );
  }
  
  return result;
}


//---------------------------------------------------------------------------//
//------------------------- sssConstructionIteration ------------------------//
//---------------------------------------------------------------------------//


pair<bool,ThRightNormalForm>
ThRightNormalForm::sssConstructionIteration
( int rank , 
  map< NF , NF >& sss_new1 ,
  map< NF , NF >& sss_checked1 ,
  const map< NF , NF >& sss_new2 ,
  const map< NF , NF >& sss_checked2 )
{
  const Permutation omega = getHalfTwistPermutation( rank );

  pair<NF,NF> pr = *sss_new1.begin( );
  NF el = pr.first;
  NF conjugator = pr.second;
  sss_checked1[el] = conjugator;
  sss_new1.erase( sss_new1.begin( ) );
  
  // 1. compute the set of simple summit conjugators 
  set<Permutation> conj = getSimpleSummitConjugators( rank , el );
  
  set<Permutation>::iterator it = conj.begin( );
  for( ; it!=conj.end( ) ; ++it ) {
    ThRightNormalForm r_mult( rank , 0 , list<Permutation>( 1 , *it ) );
    if( *it==omega ) {
      r_mult = ThRightNormalForm( rank , 1 , list<Permutation>( ) );
    }
    ThRightNormalForm l_mult = r_mult.inverse( );
    
    ThRightNormalForm res( rank );
    res = l_mult.multiply( el );
    res = res.multiply( r_mult );
    
    if( sss_new1.find( res )!=sss_new1.end( ) ||
      sss_checked1.find( res )!=sss_checked1.end( ) )
      continue;
    
    ThRightNormalForm new_conjugator = conjugator;
    new_conjugator = new_conjugator.multiply( r_mult );
    
    if( sss_new2.find( res )!=sss_new2.end( ) ||
      sss_checked2.find( res )!=sss_checked2.end( ) ) {
      ThRightNormalForm conj2( rank );
      if( sss_new2.find( res )!=sss_new2.end( ) )
        conj2 = (*sss_new2.find( res )).second;
      else
        conj2 = (*sss_checked2.find( res )).second;
      
      conj2 = conj2.multiply( new_conjugator.inverse( ) );
      
      // cout << "Found" << endl;
      return pair<bool,ThRightNormalForm>( true , conj2 );
    }
    
    // printOn( cout , ((NF)el ).second ) << " -> ";
    // printOn( cout , ((NF)res).second ) << endl;
  
    sss_new1[res] = new_conjugator;
  }
  
  return pair<bool,ThRightNormalForm>( false , ThRightNormalForm( rank ) );
}

//---------------------------------------------------------------------------//
//------------------------------ areConjugate -------------------------------//
//---------------------------------------------------------------------------//


pair<bool,ThRightNormalForm>
ThRightNormalForm::areConjugate
( const ThRightNormalForm& rep ) const
{
	if( theRank!=rep.theRank )
		return pair<bool,ThRightNormalForm>( false , ThRightNormalForm( ) );
  pair<bool,Word> result;

  // typedef pair< int , vector< Permutation > > NF;
  
  // 1. find representative of SSS1
  pair<NF,ThRightNormalForm> pr1 = findSSSRepresentative( theRank , *this );
  pair<NF,ThRightNormalForm> pr2 = findSSSRepresentative( theRank , rep   );
  NF nf1 = pr1.first;
  NF nf2 = pr2.first;
  ThRightNormalForm conj1 = pr1.second;
  ThRightNormalForm conj2 = pr2.second;

  printOn( cout , nf1.third ) << endl;
  cout << "=======================" << endl;
  printOn( cout , nf2.third ) << endl;

  {
    ThRightNormalForm tmp1 = conj1.inverse( );
    ThRightNormalForm tmp2 = tmp1.multiply( *this );
    tmp2 = tmp2.multiply( conj1 );
    if( tmp2!=nf1 ) {
      cout << "Doesn't work N 1" << endl;
      exit(1);
    }
    
    tmp1 = conj2.inverse( );
    tmp2 = tmp1.multiply( rep );
    tmp2 = tmp2.multiply( conj2 );
    if( tmp2!=nf2 ) {
      cout << "Doesn't work N 2" << endl;
      exit(1);
    }
  }
  
  if( nf1.second!=nf2.second || nf1.third.size( )!=nf2.third.size( ) )
    return pair<bool,ThRightNormalForm>( false , ThRightNormalForm( theRank ) );
  
  // cout << "(" << nf1.first << "," << nf1.second.size( ) << ")" << endl;
  
  if( nf1==nf2 ) {
    conj1 = conj1.inverse( );
    conj2 = conj2.multiply( conj1 );
    return pair<bool,ThRightNormalForm>( true , conj2 );
  }
  
  // 2. multiply by the 2-nd power of omega
  if( nf1.first<0 ) {
    int power = (1-nf1.first/2)*2;
    // int power = (-(nf1.first-1)/2)*2;
    nf1.first += power;
    nf2.first += power;
  }
  
  // 3. construct super summit sets
  map< NF , NF > sss_new1;
  map< NF , NF > sss_checked1;
  map< NF , NF > sss_new2;
  map< NF , NF > sss_checked2;
  
  sss_new1[nf1] = conj1;
  sss_new2[nf2] = conj2;
  
  for( int i=0 ; i<1000000 && sss_new1.size( ) && sss_new2.size( ) ; ++i ) {
    
    pair<bool,ThRightNormalForm> pr = sssConstructionIteration
      ( theRank , sss_new1 , sss_checked1 , sss_new2 , sss_checked2 );
    if( pr.first )
      return pair<bool,ThRightNormalForm>( true , pr.second );
    
    pr = sssConstructionIteration
      ( theRank , sss_new2 , sss_checked2 , sss_new1 , sss_checked1 );
    if( pr.first )
      return pair<bool,ThRightNormalForm>( true , pr.second.inverse( ) );
  }
  
  // cout << "Similar" << endl;
  return pair<bool,ThRightNormalForm>( true , ThRightNormalForm( theRank ) );
}


//---------------------------------------------------------------------------//
//---------------------------- computeNormalizer ----------------------------//
//---------------------------------------------------------------------------//


set<ThRightNormalForm> 
ThRightNormalForm::computeNormalizer( ) const
{
  const Permutation omega = getHalfTwistPermutation( theRank );
  
  set<ThRightNormalForm> result;
  
  map< ThRightNormalForm , ThRightNormalForm > new_states;
  map< ThRightNormalForm , ThRightNormalForm > unchecked_states;
  map< ThRightNormalForm , ThRightNormalForm > checked_states;
  
  if( theOmegaPower<0 ) {
    ThRightNormalForm rep = *this;
    rep.theOmegaPower = -theOmegaPower%2;
    cout << "Power = " << rep.theOmegaPower << endl;
    new_states[rep] = ThRightNormalForm( theRank );
  } else 
    new_states[*this] = ThRightNormalForm( theRank );
/*
  cout << endl << "  init:" << endl;
  cout << "~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~" << endl;
  cout << (*new_states.begin( )).first;
  cout << "~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~" << endl;
*/  

	while( !new_states.empty( ) ) {

    { // was unsupported by MSVC++
      // checked_states.insert( unchecked_states.begin(), unchecked_states.end() );
      map< ThRightNormalForm , ThRightNormalForm >::iterator m_it = unchecked_states.begin( );
      for( ; m_it!=unchecked_states.end( ) ; ++m_it )
        checked_states.insert( *m_it );
    }

		// process all new vertices:
    unchecked_states = new_states;
    new_states.clear( );
    
		int counter = 0;
    map< ThRightNormalForm , ThRightNormalForm >::const_iterator 
      it = unchecked_states.begin( );
    for( ; it!=unchecked_states.end( ) ; ++it ) {
      
      ThRightNormalForm cur_el = (*it).first;
      ThRightNormalForm cur_conj = (*it).second;

/*
			cout << endl << "  cur_el:" << endl;
			cout << "~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~" << endl;
			cout << cur_el;
			cout << "~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~" << endl;
*/
      // cout << "c1" << endl;
      // cout << cur_el << endl;
      // set<Permutation> conjugators = getSimpleSummitConjugators( G , cur_el );
      set<Permutation> conjugators = getSimpleConjugators( theRank , cur_el );
      // cout << "c2" << endl;
      
      for( set<Permutation>::const_iterator conj_it = conjugators.begin( ) ; conj_it!=conjugators.end( ) ; ++conj_it ) {
				
				ThRightNormalForm r_mult( theRank , 0 , list<Permutation>( 1 , *conj_it ) );
				if( *conj_it==omega )
					r_mult = ThRightNormalForm( theRank , 1 , list<Permutation>( 0 ) );
				ThRightNormalForm new_el = -r_mult * cur_el * r_mult;
				ThRightNormalForm new_conj = cur_conj * r_mult;

/*
				cout << "**********************************************" << endl;
				cout << endl << "  new_el:" << endl;
				cout << "~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~" << endl;
				cout << new_el;
				cout << "~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~" << endl;

				cout << endl << "  new_conj:" << endl;
				cout << "~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~" << endl;
				cout << new_conj;
				cout << "~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~" << endl;
*/
				
				// check whether we made a loop
				map< ThRightNormalForm , ThRightNormalForm >::const_iterator 
					fit = new_states.find( new_el );
				if( fit!=new_states.end( ) ) {
					result.insert( new_conj * -(*fit).second );
					continue;
				}
				
				fit = unchecked_states.find( new_el );
				if( fit!=unchecked_states.end( ) ) {
					result.insert( new_conj * -(*fit).second );
					continue;
				}
				
				fit = checked_states.find( new_el );
				if( fit!=checked_states.end( ) ) {
					result.insert( new_conj * -(*fit).second );
					continue;
				}
				
				cout << "Size = " << new_states.size( ) << "," << unchecked_states.size( ) << "," << checked_states.size( ) << " (" << result.size( ) << ")" << endl;
				new_states[new_el] = new_conj;
				// cout << cur_el << endl;
				// cout << new_el;
				// cout << "+++++++++++++++++++++++++++++++" << endl;

				if( counter!=result.size( ) ) {
				  ofstream of( "out.txt" );
					set<ThRightNormalForm>::iterator r_it = result.begin();
					for( ; r_it!=result.end( ) ; ++r_it ) {
						Word w = shortBraidForm( theRank , (*r_it).getShortWord() );
						of << w << endl;
					}
					counter = result.size( );
				}


				
				if( new_el.theOmegaPower<0 ) {
					cerr << "Fail" << endl;
					exit(1);
				}
      }
    }
  }


  cout << "+++++++++++++++++++++++++++++++" << endl;
  set<ThRightNormalForm>::iterator r_it = result.begin( );
	for( ; r_it!=result.end( ) ; ++r_it ) {
		cout << *r_it << endl;
		ThRightNormalForm n = -(*r_it) * (*this) * (*r_it);
		if( n!=*this ) {
			cout << "Error!" << endl;
			exit( 23 );
		}
	}


  return result;
}

