// Copyright (C) 2005 Alexander Ushakov
// Contents: Implementation of class Equation
//
// Principal Authors: Alexander Ushakov
//
// Revision History:
//

#include "Equation.h"


//---------------------------------------------------------------------------//
//--------------------------------- Equation --------------------------------//
//---------------------------------------------------------------------------//

Equation::Equation( int nGen , int nVar , const Word& eq ) :
  theNumberOfGenerators(nGen),
  theNumberOfVariables(nVar),
  theEquation(eq)
{
  
}



bool Equation::isQuadratic( ) const
{
  map< int , int > var_count;
  
  Word::const_iterator w_it = theEquation.begin( );
  for( ; w_it!=theEquation.end( ) ; ++w_it ) {
    int ag = abs( *w_it );
    if( ag>theNumberOfGenerators ) {
      var_count[ag]++;
      if( var_count[ag]==3 )
	return false;
    }
  }
  
  map< int , int >::const_iterator c_it = var_count.begin( );
  for( ; c_it!=var_count.end( ) ; ++c_it )
    if( (*c_it).second<2 )
      return false;

  return true;
}


bool Equation::isGenerator( int g ) const
{
  return abs( g )<=theNumberOfGenerators;
}


bool Equation::isVariable( int g ) const
{
  return abs( g )>theNumberOfGenerators;
}


bool Equation::trivialSolution( ) const
{
  Word image;
  Word::const_iterator w_it = theEquation.begin( );
  for( ; w_it!=theEquation.end( ) ; ++w_it )
    if( isGenerator(*w_it) )
      image.push_back( *w_it );
  
  return image.length()==0;
}


void Equation::addEquationToTransformationGraph( const Word& eq , 
						 set< Word >& chk_eq , 
						 set< Word >& new_eq , 
						 const set< int >& permutableGenerators ,
						 int x , const Word& im ) const
{
  Word image;
  Word::const_iterator eq_it = eq.begin( );
  for( ; eq_it!=eq.end( ) ; ++eq_it ) {
    int g = *eq_it;
    
    if( g==x ) {
      image.push_back( im );
    } else if( g==-x) {
      image.push_back( -im );
    } else {
      image.push_back( g );
    }
  }
  
  image = image.minimalEquivalentForm( permutableGenerators , true , true );
  // image = image.minimalEquivalentForm( permutableGenerators , false , false );
  
  if( chk_eq.find(image)!=chk_eq.end( ) || new_eq.find(image)!=new_eq.end( ) ) 
    return;
  
  static int counter = 0;
  counter++;
  new_eq.insert( image );
  if( counter%100==0 )
    cout << counter << endl;
  // cout << "=================================================" << endl;
  // cout << counter << ": " << x << " -> " << im << endl;
  // cout << image << endl;
}


pair< IntLabeledGraph , map< Word , int > > Equation::constructTransformationGraph( ) const
{
  // checked equations
  set< Word > chk_eq;
  // new equations
  set< Word > new_eq;

  set< int > permutableGenerators;
  for( int i=0 ; i<theNumberOfVariables ; ++i )
    permutableGenerators.insert( theNumberOfGenerators+i+1 );

  pair< IntLabeledGraph , map< Word , int > > result;
  // the graph
  IntLabeledGraph& fsa = result.first;
  // number of an equation
  map< Word , int >& vertex_num = result.second;
  
  new_eq.insert( theEquation.minimalEquivalentForm( permutableGenerators , true , true ) );
  while( new_eq.size( )>0 ) {
    
    // take the first unchecked equation
    Word eq = *new_eq.begin( );
    new_eq.erase( new_eq.begin( ) );
    chk_eq.insert( eq );
    
    // construct all of its descendants
    Word::const_iterator eq_it = eq.begin( );
    for( ; eq_it!=eq.end( ) ; ++eq_it ) {
      if( isGenerator( *eq_it ) )
	continue;

      // look to the left
      Word::const_iterator eq_it2 = eq_it;
      if( eq_it2!=eq.begin( ) ) {
	Word im;
	int lg = *(--eq_it2);
	if( *eq_it>0 ) {
	  im.push_back( -lg );
	  im.push_back( *eq_it );
	  addEquationToTransformationGraph( eq , chk_eq , new_eq , permutableGenerators , *eq_it , im );
	} else {
	  im.push_back( -*eq_it );
	  im.push_back( lg );
	  addEquationToTransformationGraph( eq , chk_eq , new_eq , permutableGenerators , -*eq_it , im );
	}
      }
      
      // look to the right
      eq_it2 = eq_it;
      if( ++eq_it2!=eq.end( ) ) {
	Word im;
	int lg = *eq_it2;
	if( *eq_it>0 ) {
	  im.push_back( *eq_it );
	  im.push_back( -lg );
	  addEquationToTransformationGraph( eq , chk_eq , new_eq , permutableGenerators , *eq_it , im );
	} else {
	  im.push_back( lg );
	  im.push_back( -*eq_it );
	  addEquationToTransformationGraph( eq , chk_eq , new_eq , permutableGenerators , -*eq_it , im );
	}
      }
    }
  }
  
  return result;
}


ostream& operator << ( ostream& os , const Equation& eq )
{
  os << "{" 
     << eq.getTheNumberOfGenerators( ) << " , " 
     << eq.getTheNumberOfVariables( ) << " , " 
     << eq.getTheEquation( ) << "}";
  
  return os;
}
